import os
from typing import IO, Any

from peakina.cache import timed_lru_cache

from ..fetcher import Fetcher, register
from .ftp_utils import FTP_SCHEMES, dir_mtimes, ftp_mtime, ftp_open


@timed_lru_cache(maxsize=3, seconds=60)  # type: ignore [misc]
def get_mtimes_cache(**kwargs: Any) -> dict[str, dict[str, int | None]]:
    """
    This function allows to share a common _mtime_cache object between several
    FTPFetcher objects, as long as they were instanciating with the same params.
    """
    _mtimes_cache: dict[str, dict[str, int | None]] = {}
    return _mtimes_cache


@register(schemes=FTP_SCHEMES)
class FTPFetcher(Fetcher):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._mtimes_cache: dict[str, dict[str, int | None]] = get_mtimes_cache(**kwargs)

    def get_dir_mtimes(self, dirpath: str) -> dict[str, int | None]:
        if dirpath not in self._mtimes_cache:
            self._mtimes_cache[dirpath] = dir_mtimes(dirpath)
        return self._mtimes_cache[dirpath]

    def open(self, filepath: str) -> IO[bytes]:
        return ftp_open(filepath)

    def listdir(self, dirpath: str) -> list[str]:
        """
        Make use of listdir to get all mtimes of remote files at once and
        put this information in cache.
        This will save us time by avoiding to create a new FTP connection
        for each file when we'll want to get its mtime.
        """
        return list(self.get_dir_mtimes(dirpath).keys())

    def mtime(self, filepath: str) -> int | None:
        dirpath, filename = os.path.split(filepath)
        if dirpath in self._mtimes_cache:
            return self._mtimes_cache[dirpath][filename]
        else:
            return ftp_mtime(filepath)



"""
Module to add xml support
"""
from typing import Any, cast

import jq
import pandas as pd
import xmltodict

PdDatalist = list[dict[str, Any]]
PdDatadict = dict[str, list[Any]]


def transform_with_jq(data: Any, jq_filter: str) -> PdDatalist | PdDatadict:
    """Apply a jq filter on data before it's passed to a pd.DataFrame"""
    all_data: list[PdDatalist] | list[PdDatadict] | PdDatalist = jq.all(jq_filter, data)

    # If the data is already presented as a list of rows,
    # then undo the nesting caused by "multiple_output" jq option
    if len(all_data) == 1 and (
        isinstance(all_data[0], list)  # list[PdDatalist]
        # detects another valid datastructure [{col1:[value, ...], col2:[value, ...]}]
        or (
            isinstance(all_data[0], dict) and isinstance(list(all_data[0].values())[0], list)
        )  # list[PdDatadict]
    ):
        return all_data[0]
    else:
        return cast(PdDatalist, all_data)


def read_xml(
    filepath: str,
    encoding: str = "utf-8",
    preview_offset: int = 0,
    preview_nrows: int | None = None,
    filter: str | None = None,
) -> pd.DataFrame:
    data = xmltodict.parse(open(filepath).read(), encoding=encoding)
    if filter is not None:
        data = transform_with_jq(data, filter)
    if isinstance(data, list) and isinstance(preview_nrows, int):
        data = data[preview_offset : preview_nrows + preview_offset]
    return pd.DataFrame(data)



import os
from typing import IO

from ..fetcher import Fetcher, register


@register(schemes="")
class FileFetcher(Fetcher):
    def open(self, filepath: str) -> IO[str]:
        return open(filepath)

    def listdir(self, dirpath: str) -> list[str]:
        return os.listdir(dirpath)

    def mtime(self, filepath: str) -> int:
        return int(os.path.getmtime(filepath))



import ftplib
import logging
import re
import socket
import ssl
import tempfile
from contextlib import contextmanager, suppress
from datetime import datetime
from functools import partial
from ipaddress import ip_address
from os.path import basename, join
from time import sleep
from typing import IO, Any, Callable, ContextManager, Generator, cast
from urllib.parse import ParseResult, quote, unquote, urlparse

import paramiko

FTP_SCHEMES = ["ftp", "ftps", "sftp"]

FTPClient = ftplib.FTP | paramiko.SFTPClient


class FTPS(ftplib.FTP_TLS):
    ssl_version = ssl.PROTOCOL_TLSv1_2

    def connect(  # type: ignore[override]
        self, host: str, port: int | None, timeout: int = 60
    ) -> str:
        self.host = host
        self.port = port or 990
        self.timeout = timeout

        def _setup_sock() -> socket.socket:
            _sock = socket.create_connection((self.host, self.port), self.timeout)
            self.af = _sock.family
            return _sock

        try:
            self.sock = self.context.wrap_socket(_setup_sock(), server_hostname=self.host)
        except ssl.SSLError:  # pragma: no cover
            # in some cases we must fallback to:
            self.sock = _setup_sock()

        self.file = self.sock.makefile("r")
        self.welcome = self.getresp()
        return self.welcome

    def ntransfercmd(  # type: ignore[override]
        self, cmd: str, rest: str | None = None
    ) -> tuple[socket.socket, int]:
        # override ntransfercmd so it reuses the sock session, to prevent SSLEOFError.
        # cf. https://stackoverflow.com/questions/40536061/ssleoferror-on-ftps-using-python-ftplib
        conn, size = ftplib.FTP.ntransfercmd(self, cmd, rest)
        if self._prot_p:  # type: ignore[attr-defined]
            conn = self.context.wrap_socket(
                conn, server_hostname=self.host, session=self.sock.session  # type: ignore[union-attr]
            )  # this is the fix
        return conn, size

    def makepasv(self) -> tuple[str, int]:
        # override makepasv so it rewrites the dst address if the server gave a broken one.
        # Inspired by:
        # https://github.com/lavv17/lftp/blob/d67fc14d085849a6b0418bb3e912fea2e94c18d1/src/ftpclass.cc#L774
        host, port = super().makepasv()
        if (
            self.af == socket.AF_INET and ip_address(host).is_private and self.sock is not None
        ):  # pragma: no cover
            host = self.sock.getpeername()[0]
        return host, port


@contextmanager
def ftps_client(params: ParseResult) -> Generator[tuple[FTPS, str], None, None]:
    ftps = FTPS()
    try:
        ftps.connect(host=params.hostname or "", port=params.port, timeout=3)
        try:
            ftps.prot_p()
            ftps.login(user=params.username or "", passwd=params.password or "")
        except Exception as e:
            if "SSL/TLS required on the control channel" in str(e):
                # This error means we should try the other way: first login, then prot_p:
                ftps.login(user=params.username or "", passwd=params.password or "")
                ftps.prot_p()
            else:
                raise

        yield ftps, params.path

    finally:
        with suppress(Exception):
            ftps.quit()


@contextmanager
def ftp_client(params: ParseResult) -> Generator[tuple[ftplib.FTP, str], None, None]:
    port = params.port or 21
    ftp = ftplib.FTP()
    try:
        ftp.connect(host=params.hostname or "", port=port, timeout=3)
        ftp.login(user=params.username or "", passwd=params.password or "")
        yield ftp, params.path

    finally:
        with suppress(Exception):
            ftp.quit()


@contextmanager
def sftp_client(params: ParseResult) -> Generator[tuple[paramiko.SFTPClient, str], None, None]:
    port = params.port or 22
    ssh_client = paramiko.SSHClient()
    try:
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(
            hostname=params.hostname or "",
            username=params.username,
            password=params.password,
            port=port,
            timeout=3,
        )
        sftp = ssh_client.open_sftp()
        yield sftp, params.path

    finally:
        ssh_client.close()


def _urlparse(url: str) -> ParseResult:
    """
    We had several cases of ftp usernames having '#' in them which confuses urlparse.
    The # is a reserved char for 'fragment identifiers':
    https://en.wikipedia.org/wiki/Fragment_identifier
    """
    url_params = urlparse(quote(url, safe=":/@"))  # the '#' is marked as unsafe
    # unquote to get back the original bits of string
    return ParseResult(*[unquote(param) for param in url_params])


def client(url: str) -> ContextManager[tuple[FTPClient, str]]:
    parse_result = _urlparse(url)
    ftp_client_mapping: dict[
        str, Callable[[ParseResult], ContextManager[tuple[FTPClient, str]]]
    ] = {
        "ftp": ftp_client,
        "ftps": ftps_client,
        "sftp": sftp_client,
    }
    return ftp_client_mapping[parse_result.scheme](parse_result)


def retry_pasv(c: ftplib.FTP, cmd: str, *args: Any) -> Any:
    """
    Some servers accept the PASV command, but timeout when doing commands using it.
    Solution is to retry a timeout-ed command in active mode.
    """
    fun = partial(getattr(c, cmd), *args)
    try:
        return fun()
    except socket.timeout:
        c.set_pasv(False)
        return fun()


def _open(url: str) -> IO[bytes]:
    extension = url.split(".")[-1]
    ret = tempfile.NamedTemporaryFile(suffix=f".{extension}")
    with client(url) as (c, path):
        try:
            retry_pasv(cast(ftplib.FTP, c), "retrbinary", f"RETR {path}", ret.write)
        except AttributeError:
            cast(paramiko.SFTPClient, c).getfo(path, ret)
        except ftplib.error_perm as e:
            raise Exception(f"Can't open file {path}. Please make sure the file exists") from e

    ret.seek(0)
    return ret


def ftp_open(url: str, retry: int = 4) -> IO[bytes]:  # type: ignore
    for i in range(1, retry + 1):
        try:
            return _open(url)
        except (AttributeError, OSError, ftplib.error_temp) as e:
            sleep_time = 2 * i**2
            logging.getLogger(__name__).warning(f"Retry #{i}: Sleeping {sleep_time}s because {e}")
            sleep(sleep_time)


def _get_all_files(c: FTPClient, path: str) -> list[str]:
    try:
        # retry_pasv returns path + the file
        return [basename(x) for x in retry_pasv(cast(ftplib.FTP, c), "nlst", path)]
    except AttributeError:
        if not path:
            path = "."
        return cast(paramiko.SFTPClient, c).listdir(path)


def ftp_listdir(url: str) -> list[str]:
    with client(url) as (c, path):
        return _get_all_files(c, path)


def _get_mtime(c: FTPClient, path: str) -> int | None:
    """Returns timestamp of last modification"""
    try:
        mdtm = cast(ftplib.FTP, c).sendcmd("MDTM " + path)
        # mdtm-response = "213" SP time-val CRLF (e.g. '20180101203000')
        # some FTP servers response include the milliseconds (e.g. '20180101203000.123')
        mtime = re.search(r"^213 (\d+)(\.\d+)?$", mdtm)
        assert mtime is not None
        dt = datetime.strptime(mtime.group(1), "%Y%m%d%H%M%S")
        return int((dt - datetime(1970, 1, 1)).total_seconds())
    except AttributeError:
        return cast(paramiko.SFTPClient, c).stat(path).st_mtime
    except ftplib.error_perm as e:
        logging.getLogger(__name__).warning(
            f"Can't open file {path}. Please make sure the file exists: {e}"
        )
        return None


def ftp_mtime(url: str) -> int | None:
    with client(url) as (cl_ftp, path):
        return _get_mtime(cl_ftp, path)


def dir_mtimes(url: str) -> dict[str, int | None]:
    with client(url) as (cl_ftp, path):
        mtimes_dir = {}
        all_files = _get_all_files(cl_ftp, path)
        for file in all_files:
            mtimes_dir[file] = _get_mtime(cl_ftp, join(path, file))
    return mtimes_dir



from functools import wraps
from typing import Any

import geopandas as gpd


@wraps(gpd.read_file)
def read_geo_data(
    path: str, preview_offset: int = 0, preview_nrows: int | None = None, **kwargs: Any
) -> gpd.GeoDataFrame:
    if preview_nrows and not preview_offset:
        return gpd.read_file(path, rows=preview_nrows, **kwargs)
    else:
        return gpd.read_file(
            path, rows=slice(preview_offset, preview_nrows + 1 if preview_nrows else None), **kwargs
        )



"""
This module provides access to the base Fetcher class and its registry
where all the fetchers are saved.
The main purpose of this fetcher is to retrieve a file or a list of files in case of a regex
filepath by calling the right registered subclass (local, ftp...) matching the filepath scheme.
The subclasses are all declared in the `io` directory.
"""
import fnmatch
import os
import re
from abc import ABCMeta, abstractmethod
from enum import Enum
from typing import IO, Any, Callable, Pattern, TypeVar
from urllib.parse import urlparse

from peakina.helpers import mdtm_to_string

F = TypeVar("F", bound=type["Fetcher"])


class MatchEnum(str, Enum):
    REGEX = "regex"
    GLOB = "glob"


def register(schemes: str | list[str]) -> Callable[[F], F]:
    if isinstance(schemes, str):
        schemes = [schemes]

    def f(cls: F) -> F:
        for scheme in schemes:
            cls.registry[scheme] = cls
        return cls

    return f


class Fetcher(metaclass=ABCMeta):
    """
    Base class used to:
     - list the files in a directory
     - retrieve the last modification time of a file
     - retrieve one (or many in case of a regex) TextIO from a path
    This class is used by calling `get_fetcher`, which reads the scheme of the path
    ('ftp', 's3', ...) and redirects to the right fetcher.
    All the `Fetcher` subclasses need to implement basic methods in order to be used properly.
    """

    registry: dict[str, type["Fetcher"]] = {}

    def __init__(self, **kwargs: Any) -> None:
        self.extra_kwargs = kwargs

    @classmethod
    def get_fetcher(cls, filepath: str, **fetcher_kwargs: Any) -> "Fetcher":
        scheme = urlparse(filepath).scheme
        return cls.registry[scheme](**fetcher_kwargs)

    @abstractmethod
    def listdir(self, dirpath: str) -> list[str]:
        """list all the files in a directory"""

    @abstractmethod
    def open(self, filepath: str) -> IO[bytes] | IO[str]:
        """Same as builtins `open` method in text mode"""

    @abstractmethod
    def mtime(self, filepath: str) -> int | None:
        """Get last modification time of a file"""

    @staticmethod
    def is_matching(filename: str, match: MatchEnum | None, pattern: Pattern[str]) -> bool:
        if match is None:
            return bool(filename == pattern.pattern)
        elif match is MatchEnum.GLOB:
            return bool(fnmatch.fnmatch(filename, pattern.pattern))
        else:
            return bool(pattern.match(filename))

    def get_filepath_list(self, filepath: str, match: MatchEnum | None = None) -> list[str]:
        """Methods to retrieve all the pathes to open"""
        if match is None:
            return [filepath]

        dirpath, basename = os.path.split(filepath)
        pattern = re.compile(basename)
        all_filenames = self.listdir(dirpath)
        matching_filenames = [f for f in all_filenames if self.is_matching(f, match, pattern)]
        return [os.path.join(dirpath, f) for f in sorted(matching_filenames)]

    def get_str_mtime(self, filepath: str) -> str | None:
        try:
            mdtime = self.mtime(filepath)
        except (NotImplementedError, KeyError, OSError):
            mdtime = None
        return mdtm_to_string(mdtime) if mdtime else None

    def get_mtime_dict(self, dirpath: str) -> dict[str, str | None]:
        return {f: self.get_str_mtime(os.path.join(dirpath, f)) for f in self.listdir(dirpath)}


class fetch:
    """class providing shortcuts for some Fetcher operations"""

    def __init__(self, uri: str, **fetcher_kwargs: Any) -> None:
        self.uri = uri
        self.fetcher = Fetcher.get_fetcher(uri, **fetcher_kwargs)

    @property
    def scheme(self) -> str:
        return urlparse(self.uri).scheme

    @property
    def dirpath(self) -> str:
        return os.path.dirname(self.uri)

    @property
    def basename(self) -> str:
        return os.path.basename(self.uri)

    def open(self) -> IO[bytes] | IO[str]:
        return self.fetcher.open(self.uri)

    def get_str_mtime(self) -> str | None:
        return self.fetcher.get_str_mtime(self.uri)

    def get_mtime_dict(self) -> dict[str, str | None]:
        dirpath, basename = os.path.split(self.uri)
        return self.fetcher.get_mtime_dict(dirpath)



"""
Module to add csv support
"""
from functools import wraps
from typing import TYPE_CHECKING, Any, Literal

import pandas as pd

if TYPE_CHECKING:
    from os import PathLike

    FilePathOrBuffer = str | bytes | PathLike[str] | PathLike[bytes]


@wraps(pd.read_csv)
def read_csv(
    filepath_or_buffer: "FilePathOrBuffer",
    *,
    # extra `peakina` reader kwargs
    preview_offset: int = 0,
    preview_nrows: int | None = None,
    # change of default values
    on_bad_lines: Literal["error", "warn", "skip"] = "skip",  # pandas default: "error"
    encoding_errors: Literal["strict", "ignore"] = "ignore",  # pandas default: "strict"
    **kwargs: Any,
) -> pd.DataFrame:
    """
    The read_csv method is able to make a preview by reading on chunks
    """
    if preview_nrows is not None or preview_offset:
        if (skipfooter := kwargs.pop("skipfooter", None)) is None:
            skipfooter = 0

        # In case we don't have the native nrows given in kwargs, we're going
        # to use the provided preview_nrows
        # and skipfooter should be equal to 0 because of this:
        # cf : https://github.com/pandas-dev/pandas/blob/31c553f1e599e2695a33236e02511c2841ac9aa0/pandas/io/parsers/readers.py#L2209
        if (nrows := kwargs.pop("nrows", None)) is None and skipfooter == 0:
            nrows = preview_nrows

        # In case we don't have the native skiprows given in kwargs,
        # we're going to use the provided preview_offset as range(1, preview_offset + 1)
        if (skiprows := kwargs.pop("skiprows", None)) is None:
            skiprows = range(1, preview_offset + 1)

        # keep the first row 0 (as the header) and then skip everything else up to row `preview_offset`
        all_kwargs = {"skiprows": skiprows, "skipfooter": skipfooter, "nrows": nrows}
        all_kwargs.update(kwargs)

        chunks = pd.read_csv(
            filepath_or_buffer,
            on_bad_lines=on_bad_lines,
            encoding_errors=encoding_errors,
            **all_kwargs,
        )

        # if the chunksize is not in kwargs, we want to return the iterator
        if kwargs.get("chunksize") is None:
            return next(chunks) if not isinstance(chunks, pd.DataFrame) else chunks
        return chunks

    return pd.read_csv(
        filepath_or_buffer,
        on_bad_lines=on_bad_lines,
        encoding_errors=encoding_errors,
        **kwargs,
    )


def _line_count(filepath_or_buffer: "FilePathOrBuffer", encoding: str | None) -> int:
    with open(filepath_or_buffer, encoding=encoding) as f:
        lines = 0
        buf_size = 1024 * 1024
        read_f = f.read  # loop optimization

        trailing_newline: bool = False
        while True:
            buf = read_f(buf_size)
            lines += buf.count("\n")
            if not buf:
                break
            trailing_newline = buf.endswith("\n")

        if not trailing_newline:
            lines += 1

        return lines


def csv_meta(
    filepath_or_buffer: "FilePathOrBuffer", reader_kwargs: dict[str, Any]
) -> dict[str, Any]:
    total_rows = _line_count(filepath_or_buffer, reader_kwargs.get("encoding"))

    if "names" not in reader_kwargs and total_rows > 0:  # No header row
        total_rows = total_rows - 1

    if "nrows" in reader_kwargs:
        return {
            "total_rows": total_rows,
            "df_rows": reader_kwargs["nrows"],
        }

    start = 0 + reader_kwargs.get("skiprows", 0)
    end = total_rows - reader_kwargs.get("skipfooter", 0)

    preview_offset = reader_kwargs.get("preview_offset", 0)
    preview_nrows = reader_kwargs.get("preview_nrows", None)

    if preview_nrows is not None:
        return {
            "total_rows": total_rows,
            "df_rows": min(preview_nrows, max(end - start - preview_offset, 0)),
        }
    elif preview_offset:  # and `preview_nrows` is None
        return {
            "total_rows": total_rows,
            "df_rows": max(end - start - preview_offset, 0),
        }
    else:
        return {
            "total_rows": total_rows,
            "df_rows": end - start,
        }



from .csv import csv_meta, read_csv
from .excel import excel_meta, read_excel
from .geodata import read_geo_data
from .json import read_json
from .xml import read_xml

__all__ = (
    # CSV
    "read_csv",
    "csv_meta",
    # EXCEL
    "read_excel",
    "excel_meta",
    # JSON
    "read_json",
    # XML
    "read_xml",
    # GEOJSON
    "read_geo_data",
)



import tempfile
from email.utils import parsedate_to_datetime
from typing import IO, Any

import certifi
import urllib3

from ..fetcher import Fetcher, register


@register(schemes=["http", "https"])
class HttpFetcher(Fetcher):
    def __init__(self, *, verify: bool = True, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if verify:
            self.pool_manager = urllib3.PoolManager(
                cert_reqs="CERT_REQUIRED", ca_certs=certifi.where()
            )
        else:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
            self.pool_manager = urllib3.PoolManager(cert_reqs="CERT_NONE", assert_hostname=False)

    def open(self, filepath: str) -> IO[bytes]:
        r = self.pool_manager.request("GET", filepath, preload_content=False, **self.extra_kwargs)
        ret = tempfile.NamedTemporaryFile(suffix=".httptmp")
        for chunk in r.stream():
            ret.write(chunk)
        ret.seek(0)
        return ret

    def listdir(self, dirpath: str) -> list[str]:
        raise NotImplementedError

    def mtime(self, filepath: str) -> int | None:
        try:
            r = self.pool_manager.request("HEAD", filepath)
        except Exception:
            return None
        else:
            dt = parsedate_to_datetime(r.headers["last-modified"])
            return int(dt.timestamp())



"""
Module to add excel files support
"""
import logging
from functools import wraps
from typing import Any

import pandas as pd

LOGGER = logging.getLogger(__name__)


@wraps(pd.read_excel)
def read_excel(
    *args: Any,
    preview_nrows: int | None = None,
    preview_offset: int = 0,
    **kwargs: Any,
) -> pd.DataFrame:

    df_or_dict: dict[str, pd.DataFrame] | pd.DataFrame = pd.read_excel(*args, **kwargs)

    if isinstance(df_or_dict, dict):  # multiple sheets
        for sheet_name, sheet_df in df_or_dict.items():
            sheet_df["__sheet__"] = sheet_name
        df = pd.concat(list(df_or_dict.values()))
    else:
        df = df_or_dict

    if preview_nrows:
        return df[preview_offset : preview_offset + preview_nrows or 0]

    return df


def excel_meta(filepath: str, reader_kwargs: dict[str, Any]) -> dict[str, Any]:
    """
    Returns a dictionary with the meta information of the Excel file.
    """
    excel_file = pd.ExcelFile(filepath)
    sheet_names = excel_file.sheet_names

    df = read_excel(excel_file, **reader_kwargs)

    if (sheet_name := reader_kwargs.get("sheet_name")) is None:
        # multiple sheets together
        total_rows = sum(excel_file.parse(sheet_name).shape[0] for sheet_name in sheet_names)
    else:
        # single sheet
        total_rows = excel_file.parse(sheet_name).shape[0]

    return {
        "sheetnames": sheet_names,
        "df_rows": df.shape[0],
        "total_rows": total_rows,
    }



from .fetcher import Fetcher, MatchEnum, fetch
from .ftp.ftp_fetcher import FTPFetcher
from .http.http_fetcher import HttpFetcher
from .local.file_fetcher import FileFetcher
from .s3.s3_fetcher import S3Fetcher

__all__ = (
    "Fetcher",
    "MatchEnum",
    "FTPFetcher",
    "HttpFetcher",
    "FileFetcher",
    "S3Fetcher",
    "fetch",
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['peakina',
 'peakina.io',
 'peakina.io.ftp',
 'peakina.io.http',
 'peakina.io.local',
 'peakina.io.s3',
 'peakina.readers']

package_data = \
{'': ['*']}

install_requires = \
['certifi>=2022.5.18,<2023.0.0',
 'chardet>=4,<6',
 'fastparquet>=0.8.0,<1',
 'geopandas>=0.11.1,<1',
 'jq>=1.2.1,<2.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pandas>=1.4.2,<2.0.0',
 'paramiko>=2.9.2,<3.0.0',
 'pydantic>=1.9.0,<2.0.0',
 'python-slugify>=5.0.2,<7.0.0',
 'python-snappy>=0.6.0,<1',
 's3fs>=2022.1.0,<2023.0.0',
 'tables>=3.7.0,<4.0.0',
 'urllib3>=1.26.8,<2.0.0',
 'xlrd>=2.0.1,<3.0.0',
 'xmltodict>=0.12.0,<1']

setup_kwargs = {
    'name': 'peakina',
    'version': '0.9.1',
    'description': 'pandas readers on steroids (remote files, glob patterns, cache, etc.)',
    'long_description': '[![Pypi-v](https://img.shields.io/pypi/v/peakina.svg)](https://pypi.python.org/pypi/peakina)\n[![Pypi-pyversions](https://img.shields.io/pypi/pyversions/peakina.svg)](https://pypi.python.org/pypi/peakina)\n[![Pypi-l](https://img.shields.io/pypi/l/peakina.svg)](https://pypi.python.org/pypi/peakina)\n[![Pypi-wheel](https://img.shields.io/pypi/wheel/peakina.svg)](https://pypi.python.org/pypi/peakina)\n[![GitHub Actions](https://github.com/ToucanToco/peakina/workflows/CI/badge.svg)](https://github.com/ToucanToco/peakina/actions?query=workflow%3ACI)\n[![codecov](https://codecov.io/gh/ToucanToco/peakina/branch/main/graph/badge.svg)](https://codecov.io/gh/ToucanToco/peakina)\n\n# Pea Kina _aka \'Giant Panda\'_\n\nWrapper around `pandas` library, which detects separator, encoding\nand type of the file. It allows to get a group of files with a matching pattern (python or glob regex).\nIt can read both local and remote files (HTTP/HTTPS, FTP/FTPS/SFTP or S3/S3N/S3A).\n\nThe supported file types are `csv`, `excel`, `json`, `parquet` and `xml`.\n\n:information_source: If the desired type is not yet supported, feel free to open an issue or to directly open a PR with the code !\n\nPlease, read the [documentation](https://doc-peakina.toucantoco.com) for more information\n\n# Installation\n\n`pip install peakina`\n\n# Usage\nConsidering a file `file.csv`\n```\na;b\n0;0\n0;1\n```\n\nJust type\n```python\n>>> import peakina as pk\n>>> pk.read_pandas(\'file.csv\')\n   a  b\n0  0  0\n1  0  1\n```\n\nOr files on a FTPS server:\n- my_data_2015.csv\n- my_data_2016.csv\n- my_data_2017.csv\n- my_data_2018.csv\n\nYou can just type\n\n```python\n>>> pk.read_pandas(\'ftps://<path>/my_data_\\\\d{4}\\\\.csv$\', match=\'regex\', dtype={\'a\': \'str\'})\n    a   b     __filename__\n0  \'0\'  0  \'my_data_2015.csv\'\n1  \'0\'  1  \'my_data_2015.csv\'\n2  \'1\'  0  \'my_data_2016.csv\'\n3  \'1\'  1  \'my_data_2016.csv\'\n4  \'3\'  0  \'my_data_2017.csv\'\n5  \'3\'  1  \'my_data_2017.csv\'\n6  \'4\'  0  \'my_data_2018.csv\'\n7  \'4\'  1  \'my_data_2018.csv\'\n```\n\n## Using cache\n\nYou may want to keep the last result in cache, to avoid downloading and extracting the file if it didn\'t change:\n\n```python\n>>> from peakina.cache import Cache\n>>> cache = Cache.get_cache(\'memory\')  # in-memory cache\n>>> df = pk.read_pandas(\'file.csv\', expire=3600, cache=cache)\n```\n\nIn this example, the resulting dataframe will be fetched from the cache, unless `file.csv` modification time has changed on disk, or unless the cache is older than 1 hour.\n\nFor persistent caching, use: `cache = Cache.get_cache(\'hdf\', cache_dir=\'/tmp\')`\n\n\n## Use only downloading feature\n\nIf you just want to download a file, without converting it to a pandas dataframe:\n\n```python\n>>> uri = \'https://i.imgur.com/V9x88.jpg\'\n>>> f = pk.fetch(uri)\n>>> f.get_str_mtime()\n\'2012-11-04T17:27:14Z\'\n>>> with f.open() as stream:\n...     print(\'Image size:\', len(stream.read()), \'bytes\')\n...\nImage size: 60284 bytes\n```\n\n## Installation on macOS M1 chipset\n\n## install everything\n```console\nbrew install hdf5 snappy\nHDF5_DIR="/opt/homebrew/Cellar/hdf5/1.12.1/" CPPFLAGS="-I/opt/homebrew/Cellar/snappy/1.1.9/include -L/opt/homebrew/Cellar/snappy/1.1.9/lib" poetry install\n```\n\nFor more details, here is what is needed:\n\n### install pytables\n```console\nbrew install hdf5\nHDF5_DIR="/opt/homebrew/Cellar/hdf5/1.12.1/" poetry run pip install tables\n```\n\n### install python-snappy\n```console\nbrew install snappy\nCPPFLAGS="-I/opt/homebrew/Cellar/snappy/1.1.9/include -L/opt/homebrew/Cellar/snappy/1.1.9/lib" poetry run pip install python-snappy\n```\n',
    'author': 'Toucan Toco',
    'author_email': 'dev@toucantoco.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ToucanToco/peakina',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Module to enhance pandas.read_json with JQ filter
"""
import json
from functools import wraps
from typing import TYPE_CHECKING, Any

import jq
import pandas as pd

if TYPE_CHECKING:
    from os import PathLike

    FilePathOrBuffer = str | bytes | PathLike[str] | PathLike[bytes]


def transform_with_jq(json_input: str, jq_filter: str) -> str:
    """Apply a jq filter on raw json input, outputs raw json (may be on several lines)"""
    return jq.text(jq_filter, json.loads(json_input))  # type: ignore[no-any-return]


@wraps(pd.read_json)
def read_json(
    path_or_buf: "FilePathOrBuffer",
    encoding: str = "utf-8",
    filter: str | None = None,
    preview_offset: int = 0,
    preview_nrows: int | None = None,
    *args: Any,
    **kwargs: Any,
) -> pd.DataFrame:
    if filter is None:
        filter = "."

    with open(path_or_buf, encoding=encoding) as f:
        path_or_buf = transform_with_jq(f.read(), filter)

    # In case we don't have the native nrows given in kwargs, we're going
    # to use the provided preview_nrows
    if (nrows := kwargs.get("nrows", preview_nrows)) is not None:
        if kwargs.get("lines") and kwargs.get("lines") is True:
            # cf: https://github.com/pandas-dev/pandas/blob/main/pandas/io/json/_json.py#L671
            kwargs["nrows"] = nrows
        else:
            data = json.loads(path_or_buf)
            if isinstance(data, list):
                path_or_buf = json.dumps(
                    data[preview_offset : nrows + preview_offset]
                )  # pragma: no cover

    return pd.read_json(path_or_buf, encoding=encoding, *args, **kwargs)
